import numpy as np
from .geometry_solver import GeometrySolver
from typing import TypedDict
from .solvers_abc import ABCSolver

class data(TypedDict):
    q_fuse: float
    q_sh: float
    V: float
    corpus_material: np.ndarray
    explosive_material: np.ndarray
    belt_material: np.ndarray
    gun: dict
    shell_size: dict
    corpus_coord: np.ndarray
    fuse_coord: np.ndarray
    explosive_coord: np.ndarray
    belt_coord: np.ndarray

class OFSMCCSolver:
    name = 'mcc_solver'

    flag = True

    _geo_solver = GeometrySolver()

    preprocessed_data: data = dict(
        q_fuse=None,
        q_sh=None,
        V=None,
        L=None,
        corpus_material=None,
        explosive_material=None,
        belt_material=None,
        gun=None,
        shell_size=None,
        corpus_coord=None,
        fuse_coord=None,
        explosive_coord=None,
        belt_coord=None
    )

    def preprocessor(self, data: dict, global_state: dict):

        self.preprocessed_data['q_fuse'] = data['fuse']['q']
        self.preprocessed_data['q_sh'] = data['initial_cond']['q']
        self.preprocessed_data['V'] = data['initial_cond']['V']
        self.preprocessed_data['corpus_material'] = data['materials']['corpus']
        self.preprocessed_data['explosive_material'] = data['materials']['explosive']
        self.preprocessed_data['belt_material'] = data['materials']['belt']
        self.preprocessed_data['gun'] = data['gun_char']
        self.preprocessed_data['shell_size'] = data['shell_size']

        if self.flag:
            self.flag = False
            self._geo_solver.preprocessor(data, global_state)
            self._geo_solver.run(data, global_state)
        else:
            self._geo_solver.run(data, global_state)

        geometry_state = global_state[GeometrySolver.name]
        self.preprocessed_data['corpus_coord'] = geometry_state['corpus_coord']
        self.preprocessed_data['fuse_coord'] = geometry_state['fuse_coord']
        self.preprocessed_data['explosive_coord'] = geometry_state['explosive_coord']
        self.preprocessed_data['belt_coord'] = geometry_state['belt_coord']

    @staticmethod
    # Считаем основные параметры для конуса
    def get_params_conus(r1, r2, x1, x2, rho):
        if r1 < r2:
            r1, r2 = r2, r1
            x1, x2 = x2, x1

        hi = min(x1, x2)
        h = max(x1,  x2) - hi

        lambd = r2 / r1
        # Положение цм конуса
        beta = (1 + 2 * lambd + 3 * lambd ** 2) / (4 * (1 + lambd + lambd ** 2))
        ksi = beta * h
        xc = ksi + hi

        # Объём конуса
        V =  (np.pi / 3) * (r1 ** 2 + r1 * r2 + r2 ** 2) * h
        # Масса
        q = V * rho
        # Статический момент
        S = q * xc

        # Осевой момент конуса
        mu = 0.3 * (1 + lambd + lambd ** 2 + lambd ** 3 + lambd ** 4) / (1 + lambd + lambd ** 2)
        r_xc = r1 + (r2 - r1) / (x2 - x1) * (xc - x1)
        Av = mu * V * r_xc ** 2
        A = Av * rho

        # Экваториальный момент конуса относительно донного среза (B')
        vu = (3 * (1 + lambd) ** 4 + 4 * lambd ** 2) / (80 * (1 + lambd + lambd ** 2 ) ** 2)
        Bv = 0.5 * Av + V * (vu * h** 2 + xc ** 2)
        B_sh = Bv * rho

        # Левая часть статического момента (без плотности)
        res_params = {'h': h, 'V': V, 'S': S, 'q': q, 'xc': xc, 'r_xc': r_xc, 'A': A, 'B_sh': B_sh}
        return res_params

    @staticmethod
    # Считаем основные параметры для цилиндра
    def get_params_cyl(r, x1, x2, rho):
        h = abs(x1 - x2)
        lambd = 1
        beta = 0.5
        # Положение цм цилиндра
        hi = min(x1, x2)
        ksi = beta * h
        xc = ksi + hi

        # Объём цилиндра
        V = np.pi * r ** 2 * h
        # Масса
        q = V * rho
        # Статический момент
        S = q * xc
        # Осевой момент цилиндра
        mu = 0.5
        r_xc = r
        Av = mu * V * r_xc ** 2
        A = Av * rho
        # Экваториальный момент цилиндра относительно донного среза (B')
        vu = (3 * (1 + lambd) ** 4 + 4 * lambd ** 2) / (80 * (1 + lambd + lambd ** 2) ** 2)
        Bv = 0.5 * Av + V * (vu * h ** 2 + xc ** 2)
        B_sh = Bv * rho
        res_params = {'h': h, 'V': V, 'S': S, 'q': q, 'xc': xc, 'r_xc': r_xc, 'A': A, 'B_sh': B_sh}

        return res_params

    # Расчёт МЦХ корпуса
    def solve_corp(self, p_data):
        rho = p_data['corpus_material']['rho']
        all_coord = p_data['corpus_coord']
        # Выкидываем узлы под поясок
        corp_coord = np.delete(all_coord, [4, 5], axis=1)

        x_coord = corp_coord[0]
        y_coord = corp_coord[1]
        count_nodes = len(x_coord)
        V, S, q, A, B_sh = 0, 0, 0, 0, 0

        for i in range(1, count_nodes):
            if x_coord[i] == x_coord[i - 1]:
                continue

            elif y_coord[i] == y_coord[i - 1]:
                cyl = self.get_params_cyl(r=y_coord[i], x1=x_coord[i], x2=x_coord[i - 1], rho=rho)

                if x_coord[i] > x_coord[i - 1]:
                    V += cyl['V']
                    S += cyl['S']
                    q += cyl['q']
                    A += cyl['A']
                    B_sh += cyl['B_sh']
                    # print('+Цилиндр', cyl['q'])
                else:
                    V -= cyl['V']
                    S -= cyl['S']
                    q -= cyl['q']
                    A -= cyl['A']
                    B_sh -= cyl['B_sh']
                    # print('-Цилиндр', cyl['q'])
            else:
                konus = self.get_params_conus(r1=y_coord[i], r2=y_coord[i - 1], x1=x_coord[i], x2=x_coord[i - 1],
                                          rho=rho)
                if x_coord[i] > x_coord[i - 1]:
                    V += konus['V']
                    S += konus['S']
                    q += konus['q']
                    A += konus['A']
                    B_sh += konus['B_sh']
                    # print('+Конус', konus['q'])
                else:
                    V -= konus['V']
                    S -= konus['S']
                    q -= konus['q']
                    A -= konus['A']
                    B_sh -= konus['B_sh']
                    # print('-Конус', konus['q'])


        # Расчёт параметров канавки под ВП
        vp_k1 = self.get_params_conus(r1=all_coord[1, 4], r2=corp_coord[1, 3], x1=corp_coord[0, 3], x2=all_coord[0, 4], rho=rho)
        vp_c1 = self.get_params_cyl(r=corp_coord[1, 3], x1=corp_coord[0, 3],
                                    x2=corp_coord[0, 3] + (corp_coord[0, 4]-corp_coord[0, 3])/2, rho=rho)
        vp_c2 = self.get_params_cyl(r=all_coord[1, 4], x1=all_coord[0, 4],
                                    x2=all_coord[0, 4] + (all_coord[0, 5] - all_coord[0, 4]) / 2, rho=rho)

        V -= 2 * (vp_k1['V'] + vp_c1['V'] - vp_c2['V'])
        S -= 2 * (vp_k1['S'] + vp_c1['S'] - vp_c2['S'])
        q -= 2 * (vp_k1['q'] + vp_c1['q'] - vp_c2['q'])
        A -= 2 * (vp_k1['A'] + vp_c1['A'] - vp_c2['A'])
        B_sh -= 2 * (vp_k1['B_sh'] + vp_c1['B_sh'] - vp_c2['B_sh'])

        self.q_grove = 2 * (vp_k1['q'] + vp_c1['q'] - vp_c2['q'])


        # Вывод результатов
        result_params = {'V': V, 'S': S, 'q': q, 'A': A, 'B_sh': B_sh}

        return result_params

    # Расчёт МЦХ взрывателя
    def solve_fuse(self, p_data):
        q = p_data['q_fuse']
        fuse_coord = p_data['fuse_coord']
        x_coord = fuse_coord[0]
        y_coord = fuse_coord[1]

        count_nodes = len(x_coord)
        V, S, A, B_sh  = 0, 0, 0, 0
        # Определяем сначала только объём взрывателя
        for i in range(1, count_nodes - 1):
            if x_coord[i] == x_coord[i - 1]:
                continue
            elif y_coord[i] == y_coord[i - 1]:
                cyl = self.get_params_cyl(r=y_coord[i], x1=x_coord[i], x2=x_coord[i - 1], rho=0)
                V += cyl['V']
            else:
                konus = self.get_params_conus(r1=y_coord[i], r2=y_coord[i - 1], x1=x_coord[i], x2=x_coord[i - 1], rho=0)
                V += konus['V']
        try:
            rho = q / V

        except ZeroDivisionError:
            pass

        # Определяем все остальные характеристики

        for i in range(1, count_nodes - 1):
            if x_coord[i] == x_coord[i - 1]:
                continue
            elif y_coord[i] == y_coord[i - 1]:
                cyl = self.get_params_cyl(r=y_coord[i], x1=x_coord[i], x2=x_coord[i - 1], rho=rho)
                S += cyl['S']
                A += cyl['A']
                B_sh += cyl['B_sh']
            else:
                konus = self.get_params_conus(r1=y_coord[i], r2=y_coord[i - 1], x1=x_coord[i], x2=x_coord[i - 1], rho=rho)
                S += konus['S']
                A += konus['A']
                B_sh += konus['B_sh']

        result_params = {'V': V, 'S': S, 'q': q, 'A': A, 'B_sh': B_sh}

        return result_params

    # Расчёт МЦХ ВВ
    def solve_explosive(self, p_data):
        rho = p_data['explosive_material']['rho']
        coord = p_data['explosive_coord']
        x_coord = coord[0]
        y_coord = coord[1]
        count_nodes = len(x_coord)
        V, S, q, A, B_sh = 0, 0, 0, 0, 0

        for i in range(1, count_nodes - 1):

            if x_coord[i] == x_coord[i - 1]:
                continue

            elif y_coord[i] == y_coord[i - 1]:
                if x_coord[i] > x_coord[i - 1]:
                    # Параметры цилиндрической части ВВ
                    cyl = self.get_params_cyl(r=y_coord[i], x1=x_coord[i], x2=x_coord[i - 1], rho=rho)
                    V += cyl['V']
                    S += cyl['S']
                    q += cyl['q']
                    A += cyl['A']
                    B_sh += cyl['B_sh']
                else:
                    # Параметры очка под дно взрывателя
                    och = self.get_params_cyl(r=y_coord[i], x1=x_coord[i], x2=x_coord[i - 1], rho=rho)
                    V -= och['V']
                    S -= och['S']
                    q -= och['q']
                    A -= och['A']
                    B_sh -= och['B_sh']
            else:
                # Параметры конусной части ВВ
                konus = self.get_params_conus(r1=y_coord[i], r2=y_coord[i - 1], x1=x_coord[i], x2=x_coord[i - 1], rho=rho)
                V += konus['V']
                S += konus['S']
                q += konus['q']
                A += konus['A']
                B_sh += konus['B_sh']

        result_params = {'V': V, 'S': S, 'q': q, 'A': A, 'B_sh': B_sh}
        return result_params

    # Расчёт МЦХ ВП
    def solve_belt(self, p_data):
        rho = p_data['belt_material']['rho']
        coord = p_data['belt_coord']

        # Ставим поясок на ось и сдвигаем до упора чтобы не вычитать объёмы
        x_coord = coord[0]
        y_coord = coord[1]

        ind = 1
        V_vp, S_vp, q_vp, A_vp, B_sh_vp = 0, 0, 0, 0, 0

        while x_coord[ind] != x_coord[-1]:

            if x_coord[ind] == x_coord[ind-1]:
                ind += 1

            elif y_coord[ind] == y_coord[ind-1]:
                c_outside =self.get_params_cyl(r=y_coord[ind], x1=x_coord[ind], x2=x_coord[ind-1], rho=rho)
                c_inside = self.get_params_cyl(r=y_coord[0], x1=x_coord[ind], x2=x_coord[ind-1], rho=rho)
                V_vp += c_outside['V'] - c_inside['V']
                S_vp += c_outside['S'] - c_inside['S']
                q_vp += c_outside['q'] - c_inside['q']
                A_vp += c_outside['A'] - c_inside['A']
                B_sh_vp += c_outside['B_sh'] - c_inside['B_sh']
                ind += 1

            else:
                k_outside = self.get_params_conus(r1=y_coord[ind], r2=y_coord[ind-1], x1=x_coord[ind], x2=x_coord[ind-1], rho=rho)
                k_inside = self.get_params_cyl(r=y_coord[0], x1=x_coord[ind], x2=x_coord[ind-1], rho=rho)
                V_vp += k_outside['V'] - k_inside['V']
                S_vp += k_outside['S'] - k_inside['S']
                q_vp += k_outside['q'] - k_inside['q']
                A_vp += k_outside['A'] - k_inside['A']
                B_sh_vp += k_outside['B_sh'] - k_inside['B_sh']
                ind += 1

        result_params = {'V': V_vp, 'S': S_vp, 'q': q_vp, 'A': A_vp, 'B_sh': B_sh_vp}

        return result_params

    # Расчёт гироскоп. коэфф. на вылете
    def gyro_stab(self, q, A, B, mu, xc, p_data):
        km = np.array([0.97, 0.98, 1.0, 1.03, 1.06, 1.07, 1.06, 1.05, 1.04, 1.03, 1.02, 1.01, 1.0, 0.99, 0.98, 0.97,
                       0.96, 0.95, 0.94, 0.93, 0.92, 0.91, 0.9, 0.89, 0.88])

        mach = np.array([0.762, 0.821, 0.88, 0.938, 0.997, 1.173, 1.232, 1.291,1.349, 1.467, 1.525, 1.643, 1.701, 1.76,
                         1.877, 1.995, 2.053, 2.171, 2.23, 2.347, 2.491, 2.64, 3.08, 3.227, 3.374])

        a = 340.8
        V_mach = p_data['V'] / a
        Km = np.interp(V_mach, mach, km) * 1e-3
        eta_k = p_data['gun']['eta_k']


        d = p_data['gun']['d']
        fi = 0.57
        sizes = p_data['shell_size']
        L_corp = sizes['R1'] + sizes['R2'] + sizes['R3'] + sizes['R4'] + sizes['R6'] + sizes['R7'] + sizes['R8']
        L = L_corp + sizes['R27']

        Hg = L - p_data['corpus_coord'][0, 11]

        Cq = (q / d**3) * 1e-3
        h0 = p_data['corpus_coord'][0, 11] - xc
        h1 = fi * Hg - 0.16 * d

        sizes = p_data['shell_size']
        L_corp = sizes['R1'] + sizes['R2'] + sizes['R3'] + sizes['R4'] + sizes['R6'] + sizes['R7'] + sizes['R8']
        L = L_corp + sizes['R27']

        h = h0 + h1

        if L <= 4.5 * d:
            sqrt_Ld = 1
        else:
            sqrt_Ld = np.sqrt(L / (4.5 * d))

        sigma_0 = np.sqrt(1 - (h/d) * (eta_k**2/(mu*Cq)) * (B/A) * (4/np.pi**2) * Km * sqrt_Ld)

        return sigma_0, h

    # Коррекция массы снаряда
    def correct_mass(self, curr_q, p_data):
        q1 = curr_q
        q = p_data['q_sh']
        ro1 = p_data['corpus_material']['rho']
        ro2 = p_data['explosive_material']['rho']
        R = p_data['corpus_coord'][1, 9]
        R6 = p_data['corpus_coord'][0, 9] - p_data['corpus_coord'][0, 8]

        ind = np.searchsorted(p_data['explosive_coord'][0], p_data['corpus_coord'][0, 9])

        r = np.interp(p_data['corpus_coord'][0, 9], [p_data['explosive_coord'][0, ind - 1], p_data['explosive_coord'][0, ind]],
                      [p_data['explosive_coord'][1, ind - 1], p_data['explosive_coord'][1, ind]])

        delta_A = (q - q1) / (np.pi * ((R ** 2 - r ** 2) * ro1 + r ** 2 * ro2))
        new_size = R6 + delta_A

        return new_size

    # Определяем параметры для будущего расчёта
    def get_params_for_kontur(self, data, state, xc):
        sizes = data['shell_size']

        # ng = 3 #Количество разбиения ГЧ
        # nh = ng + 1 #Количество участков головной и цилиндрической частей
        # nk = 5  #Общее количество участков на снаряде
        # nfl = 2 # Вид притупления
        # yint = 0 # Высота носика (0 потому что задаётся отдельным примитивом)
        # d = sizes['d']  # Калибра снаряда
        # hb = sizes['R17'] - (d / 2)   # Высота пояска
        # xct = state['geometry_solver']['L_all'] - xc # Координата центра тяжести (ОТ НОСИКА !!!)

        ng = 3 #Количество разбиения ГЧ
        nh = ng + 1 #Количество участков головной и цилиндрической частей
        nk = 5  #Общее количество участков на снаряде
        nfl = 2 # Вид притупления
        yint = 0 # Высота носика (0 потому что задаётся отдельным примитивом)
        d = sizes['d']  # Калибра снаряда
        hb = sizes['R17'] - (d / 2)   # Высота пояска
        xct = state['geometry_solver']['L_all'] - xc # Координата центра тяжести (ОТ НОСИКА !!!)

        R_d = d / 2
        # Колпачок взрывателя

        fuse_nose = {'type': 1, 'x1': 0., 'x2': sizes['R29'], 'r1': sizes['R28'], 'r2': sizes['R28'], 'R': 0., 'n': 0.}

        # Конус взрывателя
        fuse_korp = {'type': 2, 'x1': sizes['R29'], 'x2': sizes['R27'], 'r1': sizes['R28'], 'r2': sizes['R9'], 'R': 0., 'n': 0.}

        # Оживало снаряда
        h_oj = sizes['R27'] + sizes['R8']
        R_oj = state['geometry_solver']['Roj']
        shell_ojive = {'type': 4, 'x1': sizes['R27'], 'x2': h_oj, 'r1': sizes['R9'], 'r2': R_d, 'R': R_oj, 'n': 0.}


        # Цилиндр снаряда (до зпч)
        h_cyl = h_oj + sizes['R7'] + sizes['R6'] + sizes['R4'] + sizes['R3'] + sizes['R2']
        cyl_shell = {'type': 1, 'x1': h_oj, 'x2': h_cyl, 'r1': R_d, 'r2': R_d, 'R': 0., 'n': 0.}


        # Конус ЗПЧ
        r_zp = R_d - np.tan(sizes['R23']) * sizes['R1']
        zp_konus = {'type': 2, 'x1': h_cyl, 'x2': h_cyl + sizes['R1'], 'r1': R_d, 'r2': r_zp, 'R': 0., 'n': 0.}
        geometry_data = [fuse_nose, fuse_korp, shell_ojive, cyl_shell, zp_konus]

        result = {'ng': ng, 'nh': nh, 'nk': nk, 'nfl': nfl, 'yint': yint, 'd': d, 'hb': hb, 'xct': xct,
                  'geometry_data': geometry_data}

        return result





    # Запуск расчёта
    def run(self, data: dict, state: dict):

        settings = data['settings']['mcc']
        correct_mass_status = settings['correct_mass_status']
        # correct_mass_tol = settings['correct_mass_tol']

        p_data = self.preprocessed_data

        corpus_params = self.solve_corp(p_data)
        explosive_params = self.solve_explosive(p_data)
        belt_params = self.solve_belt(p_data)
        fuse_params = self.solve_fuse(p_data)

        # Рассчитанная масса
        q = corpus_params['q'] + explosive_params['q'] + belt_params['q'] + fuse_params['q']

        # Коррекция массы

        # Получаем новое значение R6, округляем по максимуму, до 4 знакам в мм. Дальше просто нет физического смысла
        R6 = round(self.correct_mass(q, p_data), 4)

        # Пока убираем возможность задавать точность коррекции, потому что легко свалиться в рекурсию
        # Массу снаряда подгоняем по 2 знаку после запятой
        if abs(q - p_data['q_sh']) > 1E-2 and correct_mass_status:
            self._geo_solver.preprocessed_data['R6'] = R6
            self.preprocessor(data, state)
            self.run(data, state)
        else:

            # Статический момент снаряда
            S = corpus_params['S'] + explosive_params['S'] + belt_params['S'] + fuse_params['S']

            # Осевой момент инерции снаряда
            A = corpus_params['A'] + explosive_params['A'] + belt_params['A'] + fuse_params['A']

            # Экваториальный момент снаряда относительно донного среза
            B_sh = corpus_params['B_sh'] + explosive_params['B_sh'] + belt_params['B_sh'] + fuse_params['B_sh']

            # Центр масс снаряда
            xc = S / q

            # Экваториальный момент инерции снаряда относительно его центра масс:
            B = B_sh - q * xc ** 2

            # Отношение моментов инерции
            ratio_moment = B / A

            # Коэффициент инерции снаряда
            mu = (4 * A) / (q * p_data['gun']['d']**2)
            c_q = (q / p_data['gun']['d'] ** 3) * 1e-3

            # Определяем коэфф. гироскопич. устойчивости и расст. между ЦС и ЦМ
            sigma_0, h = self.gyro_stab(q, A, B, mu, xc, p_data)

            # Коэффициент наполнения
            alpha = explosive_params['q'] / q

            res_shell = {'q': q, 'S': S, 'A': A, 'B_sh': B_sh, 'xc': xc, 'h': h, 'sigma_0': sigma_0,
                         'B': B, 'ratio_moment': ratio_moment, 'mu': mu, 'c_q': c_q, 'R6_correct': R6, 'alpha': alpha}

            corpus_params['q_grove'] = self.q_grove

            # Получаем параметры для будущего расчёта аэродинамики
            kontur = self.get_params_for_kontur(p_data, state, xc)

            results = {'shell': res_shell, 'corp': corpus_params, 'expl': explosive_params, 'belt': belt_params, 'fuse': fuse_params,
                       'kontur': kontur}

            # self.results = results
            self.flag = True

            state[OFSMCCSolver.name] = results


class APFSDMccSolver(ABCSolver):

    name = 'apfsd_mcc_solver'
    preprocessed_data = dict(
        # Активная часть
        la=None,
        da=None,
        rho_k=None,

        # Оперение
        b_kc=None,
        b_kr=None,
        l_op=None,
        h_op=None,
        n=None,
        rho_p=None,
        tr_koef=None,

        #Бал наконечник
        r_n=None,
        r_v=None,
        l_n=None,
        rho_b=None,

        # Ведущее устройство
        len=None,
        form_koef=None,
        d=None,
        rho_db=None
    )

    def _compute_kernel_mass_dim(self, p_data):

        ker_mass = 0.25 * p_data['rho_k'] * np.pi * p_data['la'] * p_data['da'] ** 2
        return dict(
            mass=ker_mass,
            mass_center=0.5 * p_data['la']
        )
    def _compute_plumage_mass_dim(self, p_data):

        H = p_data['l_op']
        L = p_data['b_kc']
        a = p_data['b_kr'] - L

        S1 = L*H
        S2 = 0.5 * a * H
        S = S1 + S2

        xc = (3 * L * S1 + (6 * L + 2 * a) * S2) / (6 * S)
        yc = H * (3 * S1 + 2 * S2) / (6 * S)

        volume = S * p_data['h_op']

        mass = volume * p_data['rho_p']

        total_volume = 0.25 * np.pi * p_data['b_kr'] * p_data['da'] ** 2
        total_volume += volume * p_data['n']
        total_volume *= p_data['tr_koef'] # Это для того чтобы учесть массу урезанную п

        return dict(
            mass=mass,
            s_op=S,
            total_volume=total_volume,
            total_mass=total_volume * p_data['rho_p'],
            mass_center_x=xc,
            mass_center_y=yc
        )

    def _compute_balltip_mass_dim(self, p_data):

        S0 = np.pi * p_data['r_n'] ** 2
        V0 = S0 * p_data['l_n'] / 3.

        l_n = p_data['r_v'] * p_data['l_n'] / p_data['r_n']
        S0 = np.pi * p_data['r_v'] ** 2
        V0 -= S0 * l_n / 3.

        mass_center = p_data['l_n'] * 2. / 3.

        mass = V0 * p_data['rho_b']

        return dict(
            mass_center=mass_center,
            mass=mass
        )

    def _compute_driving_band(self, p_data):
        la = p_data['la']

        rlen = p_data['len'] * la
        mass = 0.25 * p_data['form_koef'] * p_data['rho_db'] * rlen * np.pi * (p_data['d'] ** 2 - p_data['da'] ** 2)

        return dict(rlen=rlen, mass=mass)

    def _compute_mass_center(self, p_data, kernel, plumage, balltip):

        bt_mc = balltip['mass_center']
        bt_mass = balltip['mass']
        total_len = p_data['l_n']

        ker_mc = kernel['mass_center'] + total_len
        ker_mass = kernel['mass']
        total_len += p_data['la']

        plum_mc = plumage['mass_center_x'] + total_len
        plum_mass = plumage['total_mass']
        total_len += p_data['b_kr']

        # plum_mc = plumage['mass_center_x']
        # plum_mass = plumage['total_mass']
        # total_len = p_data['b_kr']
        #
        # ker_mc = kernel['mass_center'] + total_len
        # ker_mass = kernel['mass']
        # total_len += p_data['la']
        #
        # bt_mc = balltip['mass_center'] + total_len
        # bt_mass = balltip['mass']
        # total_len += balltip['mass_center']

        total_mass = plum_mass + ker_mass + bt_mass

        total_cm = (plum_mass * plum_mc + ker_mass * ker_mc + bt_mass * bt_mc) / total_mass

        return total_cm, total_mass, total_len
    def run(self, data: dict, global_state: dict):
        p_data = self.preprocessed_data
        kernel = self._compute_kernel_mass_dim(p_data)
        plumage = self._compute_plumage_mass_dim(p_data)
        balltip = self._compute_balltip_mass_dim(p_data)
        drive_band = self._compute_driving_band(p_data)
        total_cm, fl_mass, total_len = self._compute_mass_center(p_data, kernel, plumage, balltip)

        total_mass = fl_mass + drive_band['mass']

        global_state[APFSDMccSolver.name] = dict(
            kernel=kernel,
            plumage=plumage,
            balltip=balltip,
            drive_band=drive_band,
            total_cm=total_cm, total_mass=total_mass, fl_mass=fl_mass, total_len=total_len
        )



from typing import Sequence, Callable
import numpy as np
import numpy.random as npr
from abc import ABC, abstractmethod
import traceback
from copy import deepcopy
from shells_cae.solvers_abc import ABCSolver

class ABCOptObj(ABC):

    def __init__(self, id):
        self.id = id

class ABCAdapter(ABC):

    @abstractmethod
    def run(self, data: dict):
        pass

class ABConstraint(ABC):

    def __init__(self, constraint_type='ineq', *args, **kwargs):
        self.type = constraint_type

    @abstractmethod
    def run(self, data: dict, global_state: dict):
        pass

class OptProblem(ABC):

    def __init__(self, opt_objects: Sequence[ABCOptObj], adapters: Sequence[ABCAdapter], solvers: Sequence[ABCSolver], data: dict, predefined_state: dict):

        self.opt_objs = opt_objects
        self.adapters = adapters
        self.solvers = solvers
        self.constraints = {solver.name: [] for solver in self.solvers}

        _obj_data = {}

        self._constraint_factor = 0.

        for obj in opt_objects:
            _obj_data[obj.id] = {'obj': obj, 'var': [], 'bounds': [], 'indexes': []}

        self._obj_data = _obj_data

        global_state = {solver.name: {} for solver in solvers}

        global_state.update(predefined_state)

        self._global_state = global_state
        self._data = data
        self._index_list = []

    @property
    def data(self):
        return self._data

    @property
    def global_state(self):
        return self._global_state

    @property
    def constraint_factor(self):
        return self._constraint_factor

    def add_var_param(self, obj_id, var, lbound=-np.inf, ubound=np.inf):
        self._obj_data[obj_id]['var'].append(var)
        self._obj_data[obj_id]['bounds'].append((lbound, ubound))

    def remove_var_param(self, obj_id, var):
        if not var in self._obj_data[obj_id]['var']:
            raise KeyError('Такого варьируемого параметра у объекта нет')
        var_index = self._obj_data[obj_id]['var'].index(var)
        self._obj_data[obj_id]['var'].pop(var_index)
        self._obj_data[obj_id]['bounds'].pop(var_index)

    def add_constraint(self, solver_name, constraint: ABConstraint):
        self.constraints[solver_name].append(constraint)

    def get_x0_w_bounds(self):
        x0_list = []
        bounds_list = []
        count = 0
        for key in self._obj_data.keys():
            obj = self._obj_data[key]['obj']
            var_params = self._obj_data[key]['var']
            if not var_params:
                continue
            indexes = []
            for i, var_param in enumerate(var_params):
                x0 = getattr(obj, var_param)
                x0_bounds = self._obj_data[key]['bounds'][i]
                bounds_list.append(x0_bounds)
                x0_list.append(x0)
                indexes.append(count)
                count += 1
            self._obj_data[key]['indexes'] = indexes
        return np.array(x0_list), np.array(bounds_list)

    def set_x(self, x_array: np.ndarray):
        for key in self._obj_data.keys():
            obj = self._obj_data[key]['obj']
            var_params = self._obj_data[key]['var']
            indexes = self._obj_data[key]['indexes']

            if not var_params:
                continue

            for var_param, index in zip(var_params, indexes):
                x = x_array[index]
                setattr(obj, var_param, x)

        self.adapt()

    def adapt(self):
        for adapter in self.adapters:
            adapter.run(data=self._data)

    @abstractmethod
    def run(self):
        '''
        Например:

        for solver in self.solvers: Цикл по всем решателям
            solver.preprocessor(self._data, self._global_state) Вызов препроцессора
            solver.run(self._data, self._global_state) Запуск решателя

        :return:
        '''
        pass
        # for solver in self.solvers:
        #     solver.preprocessor(self._data, self._global_state)
        #     solver.run(self._data, self._global_state)

    @abstractmethod
    def modify_constraint(self):
        pass

    @abstractmethod
    def user_defined_objective(self):
        pass
    @abstractmethod
    def callback(self):
        # Метод вызывается на каждом удачном шаге.
        # В реализации может быть печать на экран, запись в файл и прочие операции ввода вывода для спеццифического сохранения результатов
        pass

    def __call__(self, x: np.ndarray):
        self.set_x(x)
        self.run()
        val = self.user_defined_objective()
        return val + self._constraint_factor

class SimpleConstrainedOptProblem(OptProblem):

    def run(self):
        self._constraint_factor = 0.
        for solver in self.solvers:
            solver.preprocessor(data=self._data, global_state=self._global_state)
            solver.run(self._data, self._global_state)
            solver_constraints = self.constraints[solver.name]
            if solver_constraints:
                for constraint in solver_constraints:
                    constraint_value = constraint.run(self._data, self._global_state)
                    if constraint_value > 0.:
                        self._constraint_factor = np.inf
                        return

    def modify_constraint(self):
        pass

class PenaltyConstrainedOptProblem(OptProblem):

    def __init__(self, opt_objects: Sequence[ABCOptObj], adapters: Sequence[ABCAdapter], solvers, data: dict, predefined_state: dict, r0=1., C=5.):
        super(PenaltyConstrainedOptProblem, self).__init__(opt_objects, adapters, solvers, data, predefined_state)

        self.r = r0
        self.C = C

    def modify_constraint(self):
        self.r *= self.C

    def run(self):
        eq_constaints = 0.
        ineq_constraints = 0.

        for solver in self.solvers:
            solver.preprocessor(self._data, self._global_state)
            solver.run(self._data, self._global_state)
            solver_constraints = self.constraints[solver.name]
            if solver_constraints:
                for constraint in solver_constraints:
                    constraint_value = constraint.run(self._data, self._global_state)
                    if constraint.type == 'eq':
                        eq_constaints += constraint_value ** 2
                    else:
                        ineq_constraints += max(0, constraint_value) ** 2

        self._constraint_factor = 0.5 * self.r * (eq_constaints + ineq_constraints)

class OptResult:

    def __init__(self,
                 x_history: np.ndarray,
                 f_history: np.ndarray,
                 step_mask: np.ndarray,
                 global_state_history: Sequence[dict],
                 status_code: int,
                 status_message: str,
                 error_list: Sequence[Exception]
                 ):
        self.x_history = x_history
        self.f_history = f_history
        self.step_mask = step_mask
        self.global_state_history = global_state_history
        self.status_code = status_code
        self.status_message = status_message
        self.error_list = error_list

class RandomOptimizer:

    def __init__(self, N=10, M=10, t0=0.1, R=1e-5, alpha=1.68, beta=0.68, min_delta_f=0., random_state=42):
        self.N: float = N
        self.M: float = M
        self.t0: float = t0
        self.R: float = R
        self.alpha: float = alpha
        self.beta: float = beta
        self.min_delta_f = min_delta_f
        self.random_state: int = random_state

    def _check_bounds(self, x_vec_cur: np.ndarray, bounds: np.ndarray):
        for i, x in enumerate(x_vec_cur):
            lbound = bounds[i, 0]
            ubound = bounds[i, 1]

            if x > ubound:
                x_vec_cur[i] = ubound
            if x < lbound:
                x_vec_cur[i] = lbound

    def _get_yj(self, x_cur, tk):
        """

        :param x_cur:
        :param tk:
        :return:
        """
        ksi = np.random.uniform(-1, 1, len(x_cur))
        # ksi = np.random.randn(len(x_cur))

        # if self._dropout:
        #     dropout_mask = np.random.rand(x_cur.shape[0]) > self._dropout_proba
        # else:
        #     dropout_mask = np.ones(x_cur.shape[0])

        yj = x_cur + tk * x_cur * ksi #/ np.linalg.norm(ksi)
        return yj

    def _get_zj(self, x_cur, yj):
        """

        :param x_cur:
        :param alpha:
        :param yj:
        :return:
        """
        zj = x_cur + self.alpha * (yj - x_cur)
        return zj

    def optimize(self, problem: OptProblem, print_limit=None):

        np.random.seed(self.random_state)

        global_state_history = []
        f_history = []
        mask = []
        exception_list = []
        x_history = []
        steps_total = 0
        bad_steps_cur = 0

        x0, bounds = problem.get_x0_w_bounds()
        last_x = np.copy(x0)

        last_f = np.inf

        try:
            last_f_try = problem(last_x)
            if last_f_try < last_f:
                last_f = last_f_try
                global_state_history.append(deepcopy(problem.global_state))
                f_history.append(last_f)
                x_history.append(np.copy(last_x))
                mask.append(1)
            else:
                mask.append(0)
                f_history.append(last_f)
                x_history.append((np.copy(x0)))
        except:
            mask.append(0)
            last_f = np.inf
            f_history.append(last_f)
            x_history.append((np.copy(x0)))

        tk = self.t0



        while steps_total < self.N:
            while bad_steps_cur < self.M:
                try:
                    yj = self._get_yj(last_x, tk)
                    self._check_bounds(yj, bounds)

                    f_cur = problem(yj)

                    if (f_cur < last_f) & (abs(f_cur - last_f) > self.min_delta_f):

                        zj = self._get_zj(last_x, yj)
                        self._check_bounds(zj, bounds)

                        cur_f = problem(zj)

                        if (cur_f < last_f) & (abs(cur_f - last_f) > self.min_delta_f):

                            last_x = np.copy(zj)

                            tk *= self.alpha

                            steps_total += 1
                            last_f = cur_f

                            global_state_history.append(deepcopy(problem.global_state))
                            f_history.append(last_f)
                            x_history.append(np.copy(last_x))
                            mask.append(1)
                            problem.callback()

                        else:
                            bad_steps_cur += 1
                            mask.append(0)
                            f_history.append(cur_f)
                            x_history.append((np.copy(zj)))
                    else:
                        bad_steps_cur += 1
                        mask.append(0)
                        f_history.append(f_cur)
                        x_history.append((np.copy(yj)))

                except Exception as e:

                    if print_limit is not None:
                        traceback.print_exc(limit=print_limit)
                    bad_steps_cur += 1
                    exception_list.append(e)

            if tk <= self.R:
                break
            else:
                tk *= self.beta
                bad_steps_cur = 1

        f_history = np.array(f_history)
        x_history = np.array(x_history)
        mask = np.array(mask).astype(bool)

        if last_f == np.inf:
            return OptResult(x_history=x_history,
                             f_history=f_history,
                             step_mask=mask,
                             global_state_history=global_state_history,
                             status_code=1,
                             status_message='Оптимизация завершилась неудачно, т.к. не найдено ни одного оптимума',
                             error_list=exception_list
                             )
        else:
            return OptResult(x_history=x_history,
                             f_history=f_history,
                             step_mask=mask,
                             global_state_history=global_state_history,
                             status_code=0,
                             status_message='Оптимизация завершилась удачно по критерию останова',
                             error_list=exception_list
                             )

# Алгоритм 2022 года
class SRandomOptimizer:

    def __init__(self, N=100, beta=1.0, min_delta_f=0., random_state=42):
        self.N = N
        self.beta = beta
        self.min_delta_f = min_delta_f
        self.random_state = random_state

        self.n0 = 0  # Число неудачных шагов из опорной точки
        self.supn0 = 0  # Наибольшее число неудачных шагов из какой-либо опорной точки в процессе оптимизации
        self.m = 5  # Число предшествующих шагов для адаптации поиска
        self.k = None  # Число оптимизируемых параметров
        self.rs = None
        self.ss = None
        self.MF = None
        self.M01 = None
        self.cod = None
        self.mcod = self.beta  # Начальное значение коэфф. шага
        self.KxM = None
        self.PP = None

    # Вырабаытывает случайный шаг
    def SHAG(self):
        mp = np.exp(-4.6 * ((self.n0 / self.N) ** 2 + (self.supn0 / self.N) ** 2)) / np.sqrt(self.k)
        su = 0
        sf = 0

        for i in range(self.m):
            sf = sf + self.MF[i]
            su = su + self.M01[i]

        for i in range(self.k):
            self.cod[i] = 0.0
            self.rs[i] = 0.0
            self.ss[i] = 0.0

            if np.abs(sf) > 1.0E-10:
                for j in range(self.m):
                    self.rs[i] = self.rs[i] - (j / self.m) * self.MF[j] * self.KxM[i, j]
                self.rs[i] = self.rs[i] * su * su / (self.m * sf)
                self.rs[i] = self.rs[i] * (su / self.m) * (sf / su)
                self.ss[i] = -(su / self.m) * mp * np.random.randn()
                self.cod[i] = self.rs[i] + self.ss[i]
            else:
                self.cod[i] = mp * np.random.randn()

            self.cod[i] = self.PP[i] + self.mcod * self.cod[i]

            if self.cod[i] < 0.0:
                self.cod[i] = 0.0
            if self.cod[i] > 1.0:
                self.cod[i] = 1.0

    # Определяет на каждом шаге размерность случайного подпространства
    def Sdv(self):
        for j in range(self.m - 1, 0, -1):
            self.MF[j] = self.MF[j - 1]
            self.M01[j] = self.M01[j - 1]
            for i in range(self.k):
                self.KxM[i, j] = self.KxM[i, j - 1]


    def optimize(self, problem: OptProblem, print_limit=None):
        np.random.seed(self.random_state)
        global_state_history = []
        f_history = []
        mask = []
        exception_list = []
        x_history = []
        steps_total = 0
        bad_steps_cur = 0
        Flag = False
        f_evals = 0
        f_evals_errs = 0
        x0, bounds = problem.get_x0_w_bounds()
        last_x = np.copy(x0)  # ВВП оразмеренный

        # Инициализируем переменные
        self.k = len(x0)
        # lims = np.array([bound.to_list() for bound in bounds])  # Ограничения 1 рода

        xx = x0.copy()  # Вектор в размерном виде
        last_xx = x0.copy()  # Прерыдущие значения вектора в размерном виде

        self.cod = (x0 - bounds[:, 0]) / ((bounds[:, 1]) - bounds[:, 0])  # Вектор кодовых параметров
        last_cod = self.cod.copy()  # предыдущие значения кодового вектора

        cur_f = np.inf
        last_f = np.inf

        self.PP = np.zeros(self.k)  # Сохраняет параметры опорной точки(в конце оптимизации содержит оптимальные точки)
        self.KxM = np.zeros(
            (self.k, self.m))  # Содержит приращения компонента после каждого удачного шага в последовательсти m предш. шагов
        self.MF = np.zeros(self.m)  # Содержит приращения ЦФ на каждом удачном шаге
        self.M01 = np.zeros(self.m)  # Содержит успех каждого предшествующего шага (состоит из 0 и 1)
        self.rs = np.zeros(self.k)
        self.ss = np.zeros(self.k)







        while self.n0 <= self.N:
            xx = bounds[:, 0] + (bounds[:, 1] - bounds[:, 0]) * self.cod  # Оразмеривание
            print(self.cod)
            try:
                cur_f = problem(xx)
            except:
                cur_f = np.inf

            # Проверка условий
            if (cur_f < last_f) & (abs(cur_f - last_f) > self.min_delta_f):
                # Шаг удачный
                print('Удачный')
                self.n0 = 0
                if Flag:
                    self.Sdv()
                    self.MF[0] = cur_f - last_f
                    self.M01[0] = 1
                    for i in range(self.k):
                        self.KxM[i, 0] = self.cod[i] - self.PP[i]

                Flag = True
                last_f = cur_f
                last_cod = self.cod.copy()
                last_xx = xx.copy()

                global_state_history.append(deepcopy(problem.global_state))
                f_history.append(last_f)
                x_history.append(np.copy(last_xx))
                mask.append(1)
                problem.callback()

                for i in range(self.k):
                    self.PP[i] = self.cod[i]

            else:
                # Шаг неудачный
                self.n0 += 1
                if self.n0 > self.supn0:
                    self.supn0 = self.n0
                self.Sdv()
                self.MF[0] = 0
                self.M01[0] = 0
                for i in range(self.k):
                    self.KxM[i, 0] = 0

                mask.append(0)
                f_history.append(cur_f)
                x_history.append(np.copy(xx))

            self.SHAG()

        # Выход из оптимизации
        f_history = np.array(f_history)
        x_history = np.array(x_history)
        mask = np.array(mask).astype(bool)

        if not Flag:
            return OptResult(x_history=x_history,
                             f_history=f_history,
                             step_mask=mask,
                             global_state_history=global_state_history,
                             status_code=1,
                             status_message='Оптимизация завершилась неудачно, т.к. не найдено ни одного оптимума',
                             error_list=exception_list
                             )
        else:
            return OptResult(x_history=x_history,
                             f_history=f_history,
                             step_mask=mask,
                             global_state_history=global_state_history,
                             status_code=0,
                             status_message='Оптимизация завершилась удачно по критерию останова',
                             error_list=exception_list
                             )



class ConstrainedOptimizer:

    def optimize(self, base_optimizer: RandomOptimizer, problem: OptProblem, print_limit=1,
                 n_iters=5, tol=1e-3):

        _x_history = []
        _f_history = []
        _global_state_history = []
        _mask = []
        _error_list = []

        constaint_status = 0

        res = base_optimizer.optimize(problem, print_limit)
        if res.status_code == 1:
            return res, constaint_status
        problem(res.x_history[res.step_mask][-1])
        factor = problem.constraint_factor
        if factor < tol:
            constaint_status = 1
            return res, constaint_status

        for i in range(n_iters):
            res_1 = base_optimizer.optimize(problem, print_limit)
            _x_history.extend(res_1.x_history)
            _f_history.extend(res_1.f_history)
            _global_state_history.extend(res_1.global_state_history)
            _mask.extend(res_1.step_mask)

            if res_1.status_code == 1:
                return OptResult(x_history=np.array(_x_history),
                                 f_history=np.array(_f_history),
                                 step_mask=np.array(_mask),
                                 global_state_history=_global_state_history,
                                 status_code=res.status_code,
                                 status_message=res.status_message,
                                 error_list=_error_list
                                 ), constaint_status

            # problem(res.x_history[-1])
            problem(res.x_history[res_1.step_mask][-1])
            factor = problem.constraint_factor
            problem.callback()

            if abs(factor) < tol:
                constaint_status = 1
                return OptResult(x_history=np.array(_x_history),
                                 f_history=np.array(_f_history),
                                 step_mask=np.array(_mask),
                                 global_state_history=_global_state_history,
                                 status_code=res.status_code,
                                 status_message=res.status_message,
                                 error_list=_error_list
                                 ), constaint_status
            res = res_1
            problem.modify_constraint()

        return OptResult(x_history=np.fromiter(_x_history, dtype=np.dtype((float, _x_history[0].shape))),
                         f_history=np.fromiter(map(float, _f_history), dtype=np.float),
                         step_mask=np.fromiter(_mask, dtype=int),
                         global_state_history=_global_state_history,
                         status_code=res.status_code,
                         status_message=res.status_message,
                         error_list=_error_list
                         ), constaint_status

class ParametricAnalyser(ABC):


    def compute_line(self, problem: OptProblem, post_func: Callable, var_idx=0, n_points=10):
        '''

        :param problem: Целевая функция
        :param post_func: Функция предназначенная для обработки global_state, чтобы отследить еще какую то переменную
        :param var_idx: Индекс переменной в массиве х0
        :param n_points: Кол-во точек
        :return: Массив переменной, массив функции
        '''

        x0, bounds = problem.get_x0_w_bounds()

        gl_state_line = []

        var_linspace = np.linspace(bounds[var_idx][0], bounds[var_idx][1], n_points)

        func_linspace = np.zeros_like(var_linspace)

        for i in range(n_points):
            xx = x0.copy()
            xx[var_idx] = var_linspace[i]
            problem(xx)
            gl_state_line.append(deepcopy(problem.global_state))
            func_linspace[i] = post_func(problem.global_state)

        problem.set_x(x0)


        return var_linspace, func_linspace, gl_state_line

    def compute_surface(self, problem: OptProblem, post_func: Callable, var_idx1=0, var_idx2=1, n_points=10):
        '''

        :param problem: Целевая функция
        :param post_func: Функция предназначенная для обработки global_state, чтобы отследить еще какую то переменную
        :param var_idx1: Индекс пераой переменной в массиве х0
        :param var_idx2: Индекс второй переменной в массиве х0
        :param n_points: Число точек разбиения
        :return: Сетку 1 переменной, сетку 2 переменной, сетку функции
        '''
        x0, bounds = problem.get_x0_w_bounds()
        gl_state_mesh = []

        var1_linspace = np.linspace(bounds[var_idx1][0], bounds[var_idx1][1], n_points)
        var2_linspace = np.linspace(bounds[var_idx2][0], bounds[var_idx2][1], n_points)

        mesh_var1, mesh_var2 = np.meshgrid(var1_linspace, var2_linspace)

        func_mesh = np.zeros_like(mesh_var1)

        for i in range(n_points):
            gl_state_i = []
            for j in range(n_points):
                xx = x0.copy()
                xx[var_idx1], xx[var_idx2] = mesh_var1[i, j], mesh_var2[i, j]
                problem(xx)
                gl_state_i.append(deepcopy(problem.global_state))
                func_mesh[i, j] = post_func(problem.global_state)
            gl_state_mesh.append(gl_state_i)

        problem.set_x(x0)

        return mesh_var1, mesh_var2, func_mesh, gl_state_mesh

    def compute_coefs(self, x_list, f_list, deriv_type='right'):
        '''

        :param x_list: Вектор варьируемого параметра
        :param f_list: Список целевых функций
        :param deriv_type: Тип аппроксимации производной для коэффициентов чувствительности
        :return: Абсолютный коэффициент чувствительности, относительный коэффициент чувствительности
        '''

        border = len(x_list) // 2 + 1

        if deriv_type == 'right':
            x_0 = x_list[border]
            x_1 = x_list[border + 1]

            f_0 = f_list[border]
            f_1 = f_list[border + 1]

            dx = x_1 - x_0
            df = f_1 - f_0

        elif deriv_type == 'left':
            x_0 = x_list[border]
            x_1 = x_list[border - 1]

            f_0 = f_list[border]
            f_1 = f_list[border - 1]

            dx = x_0 - x_1
            df = f_0 - f_1

        elif deriv_type == 'central':
            x_m1 = x_list[border - 1]
            x_1 = x_list[border + 1]

            f_m1 = f_list[border - 1]
            f_1 = f_list[border + 1]

            dx = x_1 - x_m1
            df = f_1 - f_m1

        aij = df / dx

        if aij != 0.:
            bij = aij * x_list[border] / f_list[border]
        else:
            bij = np.inf

        return aij, bij


        pass
    def analyze(self, problem: OptProblem, n_points=5, deriv_type='right'):
        x0, bounds = problem.get_x0_w_bounds()
        n_params = x0.shape[0]
        x_to_analyze = np.zeros((n_params, n_points, n_params))

        f_history = []
        global_state_history = []

        sense_koefs = np.zeros((n_params, 2))

        for i in range(n_params):
            param_bounds = bounds[i]
            param_linspace = np.linspace(param_bounds[0], param_bounds[1], n_points)
            for n in range(n_points):
                xx = x0.copy()
                xx[i] = param_linspace[n]
                x_to_analyze[i, n, :] = xx

        for i in range(n_params):
            param_linspace = x_to_analyze[i]
            tmp_f_history = []
            tmp_global_state_history = []
            for n in range(n_points):
                x = param_linspace[n]
                f = problem(x)
                problem.callback()
                tmp_f_history.append(f)
                tmp_global_state_history.append(deepcopy(problem.global_state))
            aij, bij = self.compute_coefs(param_linspace[:, i], tmp_f_history, deriv_type)
            sense_koefs[i, :] = aij, bij
            f_history.append(tmp_f_history)
            global_state_history.append(tmp_global_state_history)
        problem.set_x(x0)

        return x_to_analyze, f_history, global_state_history, sense_koefs













import numpy as np
from typing import TypedDict

class data(TypedDict):
    d: float
    V: float
    Hg: float
    q: float
    q_belt: float
    q_corp: float
    q_expl: float
    q_grove: float
    rho_corp: float
    rho_expl: float
    mu_c_expl: float
    sigma_t: float
    sigma_b: float
    E: float
    ground: dict
    sizes: dict
    res_geo: dict

    corpus_coord: np.ndarray
    expl_coord: np.ndarray



# Расчёт прочности ГЧ при проникании в преграду

class PenetrationHeadSolver:
    name = 'penetration_head_solver'

    preprocessed_data: data = dict(
        d=None,
        V=None,
        Hg=None,
        q=None,
        q_corp=None,
        q_expl=None,
        q_belt=None,
        q_gorve=None,
        rho_corp=None,
        rho_expl=None,
        mu_c_expl=None,
        sigma_t=None,
        sigma_b=None,
        E=None,
        ground=None,
        sizes=None,
        res_geo=None,
        corpus_coord=None,
        expl_coord=None

    )

    def preprocessor(self, data: dict, global_state: dict):
        self.preprocessed_data['d'] = data['gun_char']['d']
        self.preprocessed_data['V'] = global_state['point_mass_ebh_solver']['v_array'][-1]
        self.preprocessed_data['q'] = global_state['mcc_solver']['shell']['q']
        self.preprocessed_data['res_geo'] = global_state['geometry_solver']

        self.preprocessed_data['q_grove'] = global_state['mcc_solver']['corp']['q_grove']
        self.preprocessed_data['q_belt'] = global_state['mcc_solver']['belt']['q']
        self.preprocessed_data['q_corp'] = global_state['mcc_solver']['corp']['q']
        self.preprocessed_data['q_expl'] = global_state['mcc_solver']['expl']['q']
        self.preprocessed_data['rho_corp'] = data['materials']['corpus']['rho']
        self.preprocessed_data['rho_expl'] = data['materials']['explosive']['rho']
        self.preprocessed_data['mu_c_expl'] = data['materials']['explosive']['mu']
        self.preprocessed_data['sigma_t'] = data['materials']['corpus']['sigma_t']
        self.preprocessed_data['sigma_b'] = data['materials']['corpus']['sigma_b']
        self.preprocessed_data['E'] = data['materials']['corpus']['E']

        self.preprocessed_data['corpus_coord'] = global_state['geometry_solver']['corpus_coord']
        self.preprocessed_data['expl_coord'] = global_state['geometry_solver']['explosive_coord']
        self.preprocessed_data['sizes'] = data['shell_size']
        self.preprocessed_data['ground'] = data['ground']

    @staticmethod
    # Считаем массу конуса
    def get_params_conus(r1, r2, x1, x2, rho):
        if r1 < r2:
            r1, r2 = r2, r1
            x1, x2 = x2, x1

        hi = min(x1, x2)
        h = max(x1,  x2) - hi
        #Объём конуса
        V =  (np.pi / 3) * (r1 ** 2 + r1 * r2 + r2 ** 2) * h
        #Масса
        q = V * rho
        return q

    @staticmethod
    # Считаем массу цилиндра
    def get_params_cyl(r, x1, x2, rho):
        h = abs(x1 - x2)
        #Объём цилиндра
        V =  np.pi * r**2 * h
        #Масса
        q = V * rho
        return q

    # Определения массовых параметров корпуса в заданном сечении
    def solve_corp(self, corp_coord, rho):

        x_coord = corp_coord[0]
        y_coord = corp_coord[1]
        count_nodes = len(x_coord)
        q = 0

        for i in range(1, count_nodes):
            if x_coord[i] == x_coord[i - 1]:
                continue
            elif y_coord[i] == y_coord[i - 1]:

                q_curr = self.get_params_cyl(r=y_coord[i], x1=x_coord[i], x2=x_coord[i - 1], rho=rho)

                if x_coord[i] > x_coord[i - 1]:
                    q += q_curr
                else:
                    q -= q_curr
            else:
                q_curr = self.get_params_conus(r1=y_coord[i], r2=y_coord[i - 1], x1=x_coord[i], x2=x_coord[i - 1],
                                          rho=rho)
                if x_coord[i] > x_coord[i - 1]:
                    q += q_curr
                else:
                    q -= q_curr

        return q

    # Определения массовых параметров ВВ в заданном сечении
    def solve_explosive(self, expl_coord, rho):
        x_coord = expl_coord[0]
        y_coord = expl_coord[1]
        count_nodes = len(x_coord)
        q = 0
        for i in range(1, count_nodes - 1):

            if x_coord[i] == x_coord[i - 1]:
                continue

            elif y_coord[i] == y_coord[i - 1]:
                if x_coord[i] > x_coord[i - 1]:
                    # Параметры цилиндрической части ВВ
                    q_curr = self.get_params_cyl(r=y_coord[i], x1=x_coord[i], x2=x_coord[i - 1], rho=rho)
                    q += q_curr

                else:
                    # Параметры очка под дно взрывателя
                    q_curr = self.get_params_cyl(r=y_coord[i], x1=x_coord[i], x2=x_coord[i - 1], rho=rho)
                    q -= q_curr

            else:
                # Параметры конусной части ВВ
                q_curr = self.get_params_conus(r1=y_coord[i], r2=y_coord[i - 1], x1=x_coord[i], x2=x_coord[i - 1], rho=rho)
                q += q_curr

        return q

    # Расчёт параметров при проникании на длину ГЧ
    def gch_penetration(self, data, sec_res):

        # Характеристики преграды
        N = data['ground']['N']
        N1 = data['ground']['N1']
        N2 = data['ground']['N2']
        N3 = data['ground']['N3']
        rho = data['ground']['rho']
        rho_zv = data['ground']['rho_zv']
        a0 = data['ground']['a0']

        # Остальные данные
        q = data['q']
        H_fuse = data['sizes']['R27']
        x_sec = sec_res['x_zv'] + H_fuse
        R_sec = sec_res['Ri']
        Vc = data['V']
        # Vc = 105

        # print(f'Координаты', x_sec)
        # print(f'Радиус сечений', R_sec)

        # Разбиение ГЧ на сечения

        # Расчёт первого шага в момент встречи с преградой (kf на первом шаге считаем для взрывателя считаем с учетом взрывателя)
        kf_0 = 0.73453 - 0.49590 * np.arctan(R_sec[0] / H_fuse) + 1.23312 * (np.arctan(R_sec[0] / H_fuse)) ** 2
        S_0 = np.pi * R_sec[0] ** 2

        d0 = R_sec[0] ** 2
        Cq0 = q / d0 ** 3
        if Vc >= a0:
            Px_0 = (0.7 * N * kf_0 * rho_zv * Vc ** 2) / (1 + 0.259 * (rho_zv * N / Cq0))
        else:
            Px_0 = (0.7 * N2 * kf_0 * rho * Vc ** 2 + N3) / (1 + 0.259 * (rho * N1 / Cq0))

        V = np.zeros(10)
        Px = np.zeros(10)
        Fx = np.zeros(10)
        kf = np.zeros(10)
        S = np.zeros(10)
        t = np.zeros(10)

        V[0] = Vc
        Px[0] = Px_0
        kf[0] = kf_0
        S[0] = S_0

        for i in range(1, len(x_sec)):
            S[i] = np.pi * R_sec[i] ** 2
            di = R_sec[i] * 2
            Cqi = q / di ** 3

            dx = x_sec[i] - x_sec[i-1]
            kf[i] = 0.73453 - 0.49590 * np.arctan(R_sec[i] / x_sec[i]) + 1.23312 * (np.arctan(R_sec[i] / x_sec[i])) ** 2

            if V[i - 1] >= a0:
                k1 = 0.1 * (q + 0.33 * S[i] * rho_zv * N * di)
                k2 = 0.07 * rho_zv * N * S[i] * kf[i]
                V[i] = V[i - 1] * np.exp(- (k2 / k1) * dx)
                Px[i] = (0.7 * N * kf[i] * rho_zv * V[i] ** 2) / (1 + 0.259 * (rho_zv * N / Cqi))
                dt = (k1 / k2) * (V[i - 1] - V[i]) / (V[i - 1] * V[i])
                t[i] = t[i - 1] + dt
                Fx[i] = Px[i] * S[i]

            else:
                k1 = 0.1 * (q + 0.33 * S[i] * rho * N1 * di)
                k2 = 0.07 * S[i] * rho * kf[i] * N2
                k3 = 0.1 * S[i] * N3

                V[i] = np.sqrt(V[i - 1] ** 2 * np.exp(-2 * dx * k2 / k1) - (k3 / k2) * (1 - np.exp(-2 * dx * k2 / k1)))
                dt = k1 / np.sqrt(k2 * k3) * (np.arctan(V[i-1] * np.sqrt(k2/k3)) - np.arctan(V[i] * np.sqrt(k2/k3)))
                t[i] = t[i - 1] + dt
                Px[i] = (0.7 * N2 * kf[i] * rho * V[i] ** 2 + N3) / (1 + 0.259 * (rho * N1 / Cqi))
                Fx[i] = Px[i] * S[i]


        # Расчёт изменения скорости и давления преграды по сечениям
        # for i in range(1, len(x_sec)):
        #     S[i] = np.pi * R_sec[i] ** 2
        #     kf[i] = 0.73453 - 0.49590 * np.arctan(R_sec[i] / x_sec[i]) + 1.23312 * (np.arctan(R_sec[i] / x_sec[i])) ** 2
        #
        #     if V[i - 1] >= a0:
        #         k1 = 0.1 * (q + 0.33 * S[i] * rho_zv * N * d)
        #         k2 = 0.07 * rho_zv * N * S[i] * kf[i]
        #         V[i] = V[i - 1]  * np.exp(- (k2 / k1) * x_sec[i])
        #         Px[i] = (0.7 * N * kf[i] * rho_zv * V[i] ** 2) / (1 + 0.259 * (rho_zv * N / Cq))
        #         # print('Сверхзвук')
        #     else:
        #         k1 = 0.1 * (q + 0.33 * S[i] * rho * N1 * d)
        #         k2 = 0.07 * S[i] * rho * kf[i] * N2
        #         k3 = 0.1 * S[i] * N3
        #         V[i] = np.sqrt(V[i - 1] ** 2 * np.exp(-2 * x_sec[i] * k2 / k1) - (k3 / k2) * (1 - np.exp(-2 * x_sec[i]  * k2 / k1)))
        #         Px[i] = (0.7 * N2 * kf[i] * rho * V[i] ** 2 + N3) / (1 + 0.259 * (rho * N1 / Cq))



        results = {'x_with_fuse': x_sec, 't': t, 'V': V, 'S': S, 'kf': kf, 'Px': Px, 'Fx': Fx}

        # my_plot(x=x_sec ,y=V, title='Изменение скорости снаряда при проникании на длину ГЧ \n (преграда СПГ Vc=105)')
        # print('Давление преграды: ', Px * 1E-6)
        # print('Скорость снаряда: ', V)
        # print('Коэфф. головной части: ', kf)
        # print('Площадь: ', S)
        # print('Время проникания: ', t)

        return results





    # Определение массива сечений со значениями xi, ri, Ri
    def section_solver(self, data):
        # Формирования сечений, получение массивов xi, ri, Ri
        n_sec = 10
        xi_sec = data['corpus_coord'][0, 11:21]
        Ri_sec = data['corpus_coord'][1, 11:21]
        ri_sec = np.zeros(10)
        r_zv = np.zeros(10)
        H_shell = np.max(data['corpus_coord'][0])
        x_max_exlp = np.max(data['expl_coord'][0])
        corp_bottom = np.flip(data['corpus_coord'][:, 21:], axis=1)

        for i in range(len(xi_sec)):
            curr_x = xi_sec[i]
            ind = np.searchsorted(corp_bottom[0], curr_x)
            r = np.interp(curr_x, [corp_bottom[0, ind - 1], corp_bottom[0, ind]],
                          [corp_bottom[1, ind - 1], corp_bottom[1, ind]])
            ri_sec[i] = r

            # Храним радиусы для прочности (с занулением для сечений со взрывателем)
            if curr_x < x_max_exlp:
                r_zv[i] = r




        x_zv = H_shell - xi_sec

        result = {'n_sec': n_sec, 'x_zv': np.flip(x_zv), 'xi': np.flip(xi_sec), 'Ri': np.flip(Ri_sec),
                  'r_zv': np.flip(r_zv), 'ri': np.flip(ri_sec)}


        return result

    # Определение наседающих масс
    def press_mass_solver(self, data, sec_res):

# Постоянная наседающая масса давящая ГЧ

        # q_belt = data['q_belt']
        # q_corp = data['q_corp']
        # q_expl = data['q_expl']

        expl_coord = data['expl_coord']
        corp_coord = np.delete(data['corpus_coord'], [4, 5], axis=1)    # выкидываем координаты под поясок

        # Постоянная наседающая масса
        # Масса корпуса до ГЧ
        before_gch_corp_coord = np.delete(corp_coord, np.where(corp_coord[0] > sec_res['xi'][-1]), axis=1)
        before_gch_corp_coord = np.insert(before_gch_corp_coord, [10], [[sec_res['xi'][-1]], [sec_res['ri'][-1]]], axis=1)
        q_before_gch_corp = self.solve_corp(before_gch_corp_coord, data['rho_corp']) - data['q_grove']                      # Прибавить поясок

        # Масса ВВ до ГЧ
        # before_gch_expl_coord = np.delete(expl_coord, np.where(expl_coord[0] > sec_res['xi'][-1]), axis=1)
        # before_gch_expl_coord = np.insert(before_gch_expl_coord, [-1], [[sec_res['xi'][-1], sec_res['xi'][-1]], [sec_res['ri'][-1], 0]], axis=1)
        # q_before_gch_expl = self.solve_explosive(before_gch_expl_coord, data['rho_expl'])


# Расчёт наседающих масс по сечениям
        q_press_corp = {}
        q_press_expl = {}
        q_press_shell = {}
        q_press_im_expl = {}
        x_gch = sec_res['xi'][-1]
        r_gch = sec_res['ri'][-1]

        for i in range(len(sec_res['xi'])):
            x_sec = sec_res['xi'][i]
        # Наседающая корпуса
            # Откидываем из всех координат всё что до ГЧ
            after_gch_corp_coord = np.delete(corp_coord, np.where(corp_coord[0] < x_gch), axis=1)
            after_gch_corp_coord = np.insert(after_gch_corp_coord, [0], [[x_gch], [r_gch]], axis=1)
            after_gch_corp_coord = np.append(after_gch_corp_coord, [[x_gch], [r_gch]], axis=1)

            # Массы по сечениям
            curr_corp_coord = np.delete(after_gch_corp_coord, np.where(after_gch_corp_coord[0] > x_sec), axis=1)
            ind = curr_corp_coord[0].argmax()
            curr_corp_coord = np.insert(curr_corp_coord, [ind+1], [[sec_res['xi'][i]], [sec_res['ri'][i]]], axis=1)

            q_curr_corp = self.solve_corp(corp_coord=curr_corp_coord, rho=data['rho_corp'])
            q_press_corp[i + 1] = q_curr_corp + q_before_gch_corp
            # my_plot(curr_corp_coord, title=f'Сечение № {i + 1}')

        # Наседающая ВВ (тут считаются сечения справа-налево, а затем масса отнимается от общей массы ВВ)
            # Откидываем из всех координат всё что до ГЧ
            after_gch_expl_coord = np.delete(expl_coord, np.where(expl_coord[0] < x_gch), axis=1)
            after_gch_expl_coord = np.insert(after_gch_expl_coord, [0], [[x_gch], [r_gch]], axis=1)

            curr_expl_coord = np.delete(after_gch_expl_coord, np.where(after_gch_expl_coord[0] < x_sec), axis=1)
            curr_expl_coord = np.insert(curr_expl_coord, [0], [[sec_res['xi'][i]], [sec_res['ri'][i]]], axis=1)

            # my_plot(curr_expl_coord, title=f'Сечение № {i + 1}')

            q_curr_expl = self.solve_explosive(expl_coord=curr_expl_coord, rho=data['rho_expl'])
            q_press_expl[i + 1] = data['q_expl'] - q_curr_expl

        # Суммарная наседающая
            q_press_shell[i + 1] = q_press_corp[i + 1] + q_press_expl[i + 1]

        # Масса в воображаемом цилиндре
            n_oj_points = 10      # Число точек по которым строится оживало каморы
            r_sec = sec_res['ri'][i]
            x_expl_max = np.max(expl_coord[0])

            if x_sec > x_expl_max:
                q_curr_im_expl = 0
            else:
                # Удаляем координаты правее сечения, добавляем точки, чтобы замкнуть контур
                curr_expl_coord = np.delete(expl_coord, np.where(expl_coord[0] > x_sec), axis=1)
                # Если есть очко под взрыватель
                if (expl_coord[0, -3] != expl_coord[0, -4]) and (expl_coord[0, -3] <= x_sec <= expl_coord[0, -4]):
                    curr_expl_coord = np.insert(curr_expl_coord, [-3], [[x_sec, x_sec], [r_sec, expl_coord[1, -3]]], axis=1)
                else:
                    curr_expl_coord = np.insert(curr_expl_coord, [-1], [[x_sec, x_sec], [r_sec, 0.0]], axis=1)

                # Проверяем точку максимума по радиусу
                r_expl_max = np.max(curr_expl_coord[1])
                if r_sec >= r_expl_max:
                    q_curr_im_expl = self.solve_explosive(expl_coord=curr_expl_coord, rho=data['rho_expl'])
                else:
                    # Удаляем координаты выше сечения, добавляем точки, чтобы замкнуть контур
                    curr_expl_coord = np.delete(curr_expl_coord, np.where(curr_expl_coord[1] > r_sec), axis=1)
                    if r_sec < expl_coord[1, 1]:
                        curr_expl_coord = np.insert(curr_expl_coord, [1], [[expl_coord[0, 0]], [r_sec]], axis=1)
                        q_curr_im_expl = self.solve_explosive(expl_coord=curr_expl_coord, rho=data['rho_expl'])
                    else:
                        max_r_ind = 4 # индекс максимального радиуса по каморе
                        ind = np.searchsorted(expl_coord[1, :max_r_ind], r_sec)
                        x_interp = np.interp(r_sec, [expl_coord[1, ind-1], expl_coord[1, ind]],
                                                    [expl_coord[0, ind-1], expl_coord[0, ind]])
                        curr_expl_coord = np.insert(curr_expl_coord, [ind], [[x_interp], [r_sec]], axis=1)
                        q_curr_im_expl = self.solve_explosive(expl_coord=curr_expl_coord, rho=data['rho_expl'])

            q_press_im_expl[i + 1] = q_curr_im_expl

            # my_plot(curr_expl_coord, title=f'Сечение № {i + 1}')


        results = {'q_press_corp': q_press_corp, 'q_press_expl': q_press_expl,
                   'q_press_shell': q_press_shell, 'q_press_im_expl': q_press_im_expl}
        return results

    # Определение напряжений в головной части снаряда при проникании
    def stress_solver(self, data, pen_res, sec_res, press_mass_res):
        sec_count = len(sec_res['xi'])
        q = data['q']
        mu_c = data['mu_c_expl']
        px = pen_res['Px'][-1]
        Rx = sec_res['Ri'][-1]
        Sn = np.zeros(sec_count)
        Nn = np.zeros(sec_count)
        pc = np.zeros(sec_count)
        sigma_x = np.zeros(sec_count)
        sigma_r = np.zeros(sec_count)
        sigma_t = np.zeros(sec_count)
        sigma_pr = np.zeros(sec_count)

        for i in range(sec_count):
            Rn = sec_res['Ri'][i]
            rn = sec_res['r_zv'][i]
            qn = press_mass_res['q_press_shell'][i + 1]
            w_im_expl = press_mass_res['q_press_im_expl'][i + 1]

            # Площадь корпуса снаряда в сечении
            Sn[i] = np.pi * (Rn**2 - rn**2)

            # Сжимающая сила
            if i != 9:
                # формула для гч
                Nn[i] = px * np.pi * Rx**2 * (qn / q - 1 + Rn**2/Rx**2)
            else:
                # формула для перехода ГЧ в корпус
                Nn[i] = px * np.pi * Rx**2 * (qn / q)

            # Осевое напряжение
            sigma_x[i] = - Nn[i] / Sn[i]


            if rn!= 0.0:
                # Радиальное напряжение
                sigma_r[i] = - (mu_c / (1 - mu_c)) * px * (w_im_expl / q) * (Rx**2 / rn**2)
                # Давление снаряжения
                pc[i] = - sigma_r[i]
                # Тангенциальное напряжение
                k = Rn ** 2 / rn ** 2
                if i != 9:
                    # формула для ГЧ
                    sigma_t[i] = - 2 * px * (k / (k - 1)) + pc[i] * ((k + 1) / (k - 1))
                else:
                    # формула для перехода ГЧ в корпус
                    sigma_t[i] = pc[i] * (k + 1 / (k - 1))

            else:
                sigma_r[i] = 0.0
                pc[i] = 0.0
                sigma_t[i] = 0.0




            # Приведенное напряжение
                # Сортируем значения (от большего к меньшему)
            s = np.array([sigma_x[i], sigma_r[i], sigma_t[i]])
            s.sort(kind='quicksort')
            s = np.flip(s)
            sigma_pr[i] = 1 / np.sqrt(2) * np.sqrt((s[0] - s[1])**2 + (s[1] - s[2])**2 + (s[2] - s[0])**2)



        results = {'S': Sn, 'pc': pc, 'N': Nn, 'sigma_x': sigma_x, 'sigma_r': sigma_r,
                   'sigma_t': sigma_t, 'sigma_pr': sigma_pr}

        return results


    # Допустимое напряжение в материале корпуса снаряда
    def allowable_stress(self, data, pen_res, sigma_pr_max):
        sigma_t = data['sigma_t'] * 1E-9    # в ГПа
        E = data['E'] * 1E-9    # в ГПа
        sigma_pr_max = sigma_pr_max * 1E-9 # в ГПа
        sigma_b = data['sigma_b'] * 1E-9 # в ГПа

        if data['sigma_t'] < 1e9:
            n = 0.8
        else:
            n = 2


        # Время проникания на длину ГЧ
        tx = pen_res['t'][-1]

        eps = sigma_pr_max / (E * tx)
        A = 1 + 0.1 * (3 **  (0.22 * np.log(eps)))
        Kt = 1 + np.log(A) / (1.35 * sigma_t ** n)

        sigma_dt = Kt * sigma_t
        nz = sigma_dt / sigma_pr_max
        sigma_dop = sigma_t * Kt / 1.04

        result = {'sigma_pr_max': sigma_pr_max, 'sigma_dop': sigma_dop, 'sigma_b': sigma_b, 'nz': nz}

        return result



    # Определяем координаты для каждого сечения
    def run(self, data: dict, state: dict):
        p_data = self.preprocessed_data
        name_ground = data['ground']['name']



        sec_res = self.section_solver(p_data)
        pen_res = self.gch_penetration(p_data, sec_res)
        press_mass_res = self.press_mass_solver(p_data, sec_res)
        stress_res = self.stress_solver(p_data, pen_res, sec_res, press_mass_res)

        sigma_pr_max = np.max(stress_res['sigma_pr'])
        safe_factor_res = self.allowable_stress(p_data, pen_res, sigma_pr_max)


        # Проверка условия проникания
        N3 = p_data['ground']['N3']
        rho = p_data['ground']['rho']

        # Минимальная скорость при которой возможно проникание
        V_dop = np.sqrt(N3 / rho)

        results = {'V_dop': V_dop, 'ground': name_ground}
        results.update(pen_res)
        results.update(sec_res)
        results.update(press_mass_res)
        results.update(stress_res)
        results.update(safe_factor_res)

        state[PenetrationHeadSolver.name] = results




        # if V >= V_dop:
        #     print('Проникание в преграду возможно.')
        # else:
        #     print('ПРОНИКАНИЕ НЕВОЗМОЖНО!')
        #     return


        # my_plot(np.array([sec_res['x_zv'], stress_res['sigma_pr'] * 1e-6 ]), title='Приведенные напряжения')




# Решалы
from .geometry_solver import GeometrySolver

from .mcc_solver import OFSMCCSolver
from .mcc_solver import APFSDMccSolver

from .strength_solver import CoordSectionSolver
from .strength_solver import PressMassSolver
from .strength_solver import StressSolver
from .strength_solver import SafeFactorSolver
from .strength_solver import BottomThickSolver
from .strength_solver import DeformCylSolver
from .strength_solver import RampCorpSolver
from .strength_solver import BeltZoneSolver
from .strength_solver import APFSDStrengthSolver

from .mcdrag_solver import McDragSolver

from .eb_solver import PointMassTrajectoryHSolver
from .eb_solver import PointMassTrajectorySolver

# from .kontur_solver import KonturSolver

from .penetration import PenetrationHeadSolver

from .solvers_abc import ABCSolver


from .al_tate_solver import AlTateSolver
from .apfsd_fl_stable_solver import APFSDFlStableSolver
from .ib_solver import *

# Оптимизация
from .optimizers import optimizers



import cffi
import os
import numpy as np
import sys
from typing import Sequence, Union
from .solvers_abc import ABCSolver

def load_lib():
    HERE = os.path.dirname(__file__)
    if sys.platform.startswith('linux'):
        LIB_FILE_NAME = os.path.abspath(os.path.join(HERE, ".", "compiled", "build", "bin", "lib", "libpenetrationlib.so"))
    elif sys.platform.startswith('win32'):
        LIB_FILE_NAME = os.path.abspath(os.path.join(HERE, ".", "compiled", "build", "bin", "libpenetrationlib.dll"))
    else:
        raise Exception('Неподдерживаемая платформа')
    ffi = cffi.FFI()
    ffi.cdef(
        '''
        typedef struct material{\n
        double sigma;\n
        double rho;\n
        double C;\n
        double a;\n
        double lambda;\n
        } material;
        '''
    )
    ffi.cdef(
        '''
        typedef struct hedge{\n
        material mat;\n
        double E;\n
        double nu;\n
        double lambda1;\n
        double A;\n
        double gamma;\n
        } hedge;
        '''
    )
    ffi.cdef(
        '''
        typedef struct kernel{\n
        material mat;\n
        double c_zv;\n
        double da;\n
        double la;\n
        double vc;\n
        } kernel;
        '''
    )

    ffi.cdef(
        '''
        typedef struct pen_prob{\n
        double u_0;\n
        double d;\n
        double r_g;\n
        } pen_prob;
        '''
    )
    ffi.cdef(
        '''
        void set_hedge_material(\n
        hedge *ahedge,\n
        double sigma,\n
        double rho,\n
        double a,\n
        double lambda\n
        );
        '''
    )
    ffi.cdef(
        '''
        void set_kernel_material(\n
        kernel *akernel,\n
        double sigma,\n
        double rho,\n
        double a,\n
        double lambda\n
        );
        '''
    )

    ffi.cdef(
        '''
        void init_hedge(\n
        hedge *ahedge,\n
        double E,\n
        double nu,\n
        double lambda1\n
        );
        '''
    )
    ffi.cdef(
        '''
        void init_kernel(\n
        kernel *akernel,\n
        double c_zv,\n
        double da,\n
        double la,\n
        double vc\n
        );
        '''
    )

    ffi.cdef(
        '''
        void compute_penetration(\n
        double *y_s,\n
        pen_prob *prob,\n
        kernel *akernel,\n
        hedge *ahedge,\n
        double tstep,\n
        double tmax\n
        );
        '''
    )

    ffi.cdef(
        '''
        void dense_compute_penetration(\n
        double *y_s,\n
        int *ntsteps,\n
        pen_prob *prob,\n
        kernel *akernel,\n
        hedge *ahedge,\n
        double tstep,\n
        double tmax\n
        );
        '''
    )

    bal_lib = ffi.dlopen(LIB_FILE_NAME)
    return ffi, bal_lib

FFI, PEN_LIB = load_lib()

class AlTateSolver(ABCSolver):
    name = 'al_tate_solver'

    preprocessed_data = dict(
        hedge_sigma=None,
        hedge_rho=None,
        hedge_c0=None,
        hedge_b0=None,
        kernel_sigma=None,
        kernel_rho=None,
        kernel_c0=None,
        kernel_b0=None,

        hedge_E=None,
        hedge_nu=None,
        hedge_lambd=None,

        kernel_c_zv=None,
        kernel_da=None,
        kernel_la=None,
        kernel_vc=None,

    )

    def run(self, data: dict, global_state: dict):

        p_data = self.preprocessed_data
        kernel = FFI.new('kernel *')
        hedge = FFI.new('hedge *')
        prob = FFI.new('pen_prob *')

        al_tate_settings = data['settings'][AlTateSolver.name]

        tstep = al_tate_settings['tstep']
        tmax = al_tate_settings['tmax']
        n_tsteps = FFI.new('int *')
        n_tsteps[0] = int(tmax / tstep)

        PEN_LIB.set_hedge_material(
            hedge,
            p_data['hedge_sigma'], p_data['hedge_rho'],
            p_data['hedge_c0'], p_data['hedge_b0']
        )

        PEN_LIB.set_kernel_material(
            kernel,
            p_data['kernel_sigma'], p_data['kernel_rho'],
            p_data['kernel_c0'], p_data['kernel_b0']
        )

        PEN_LIB.init_hedge(hedge,
                           p_data['hedge_E'],
                           p_data['hedge_nu'],
                           p_data['hedge_lambd']
                           )

        PEN_LIB.init_kernel(kernel, p_data['kernel_c_zv'],
                            p_data['kernel_da'], p_data['kernel_la'], p_data['kernel_vc'])


        y_s = np.zeros((4, n_tsteps[0]), order='F', dtype=np.float64)
        y_s_ptr = FFI.cast("double*", y_s.__array_interface__['data'][0])
        PEN_LIB.dense_compute_penetration(
            y_s_ptr, n_tsteps, prob, kernel, hedge, tstep, tmax
        )

        y_s = y_s[:, :n_tsteps[0]-1]
        t_s = np.linspace(0., tstep * n_tsteps[0], n_tsteps[0] - 1)

        global_state[AlTateSolver.name] = dict(
            t_array=t_s,
            x_array=y_s[0],
            l_array=y_s[1],
            v_array=y_s[2],
            u_array=y_s[3],
            u_v_array=y_s[2] - y_s[3]
        )


from setuptools import setup

setup(name='shells_cae',
      version='0.0.19',
      description='Special CAЕ Library by sabbraxcaddabra&denchina',
      packages=['shells_cae', 'shells_cae.optimizers'],
      package_data={'shells_cae':[
            'compiled/al_tate_model/src/*.f90',
            'compiled/al_tate_model/CMakeLists.txt',
            'compiled/external_ballistics/src/*.f90',
            'compiled/external_ballistics/CMakeLists.txt',
            'compiled/internal_ballistics/src/*.f90',
            'compiled/internal_ballistics/CMakeLists.txt',
            'compiled/mcdrag/src/*.f90',
            'compiled/mcdrag/CMakeLists.txt',
            'compiled/kontur/include/*.h',
            'compiled/kontur/src/*.for',
            'compiled/kontur/src/*.c',
            'compiled/kontur/CMakeLists.txt',
            'compiled/build/bin/*.*',
            'compiled/build/bin/lib/*.so'
      ]},


      author_email='denisdoronichev@outlook.com',
      include_package_data=True,
      zip_safe=False,
      install_requires=['numpy>=1.23.4', 'cffi>=1.15.1'],
      python_requires='>=3.8'
      )


# Команда для обновления файлов сборки
# python setup.py sdist

# Отправка на сервер новой версии
# twine upload dist/*


import cffi, os, sys
import numpy
import numpy as np
from typing import TypedDict

def load_lib():

    HERE = os.path.dirname(__file__)
    if sys.platform.startswith('linux'):
        LIB_FILE_NAME = os.path.abspath(os.path.join(HERE, ".", "compiled", "build", "bin", "lib", "libkontur.so"))
    elif sys.platform.startswith('win32'):
        LIB_FILE_NAME = os.path.abspath(os.path.join(HERE, ".", "compiled", "build", "bin", "libkontur.dll"))
    else:
        raise Exception('Неподдерживаемая платформа')
    ffi = cffi.FFI()

    ffi.cdef(
        '''
        typedef struct csv_st csv;
        
        enum geometry_primitives {
          G_CYLINDR = 1,       //Цилиндр
          G_CONE = 2,          //Конус
          G_REVERSE_OGIVE = 3, //Обратное оживало
          G_OGIVE = 4,         //Оживало
          G_EXPONENT = 5,      //Экспонента
          G_PARABOLIC =6       //Парабола
        };
        
        csv *CSV_New(void);
        void CSV_SetNG(csv *p_csv, const int ng); //Кол-во участков в головной части
        void CSV_SetNH(csv *p_csv, const int nh); //Кол-во участков говной и цилиндрической
                                      // частей (ng или ng+1)
        void CSV_SetNK(csv *p_csv, const int nk); //Общее кол-во участков на снаряде
        void CSV_SetNFL(csv *p_csv, const int nfl); //Вид притупления nfl=1 - сферическое,
                                        // nfl= 2..100- другое
        void CSV_SetISIG(csv *p_csv, const int isig); //Не обязательно!
                       //Параметр учета толщины вытеснения пограничного слоя
                       //isig = 0 - учитывается (L <= 7*d) (по умолчанию)
                       //isig = 1 - не учитывается (L > 7*d)
        void CSV_SetYInt(csv *p_csv, const double yint); //Высота носика, м
        void CSV_SetDia(csv *p_csv, const double d); //Диаметр снаряда, м
        void CSV_SetHb(csv *p_csv, const double hb); //Высота пояска, м
        void CSV_SetA(csv *p_csv, const double a); //Не обязательно! Скорость звука, м/с.
                                          //(По умолчанию 340,8)
        void CSV_SetRo(csv *p_csv, const double ro); //Не обязательно! Плотность воздуха,
                                         //м/с. (По умолчанию 0,1229)
        void CSV_SetAMU(csv *p_csv, const double amu); //Не обязательно! Динамическая
                           //вязкость воздуха, Н*с/m^2 (По умолчанию 1,825E-6)
        void CSV_SetC2(csv *p_csv, const double c2); //Не обязательно! Шаг сетки по оси X, м
                                         //(По умолчанию 0,02)
        void CSV_SetXct(csv *p_csv, const double xct);//Координата центра тяжести от носика, м
        void CSV_AddGeom(csv *p_csv, const int num, const int type,
                         const double x1, const double x2,
                         const double r1, const double r2,
                         const double rd, const double exp);
        /* Добавляет/заменяет геометрический примитив
         * при замене нужно переопределить все примитивы!
         * num - Индекс примитива начиная с 0
         * type - Тип примитива (см. geometry_primitives)
         * x1, x2 - Начальная и конечная координата x от носика
         * r1, r2 - Радиус в начальной и конечной точке
         * rd - Радиус, если оживало, м
         * exp - Показатель степени, если экспонента
         */
        
        const char *CSV_CheckError(csv *p_csv); //Проверяет готовность к расчету
                                          //Возвращает описание ошибки или NULL
        
        void CSV_Solve(csv *p_csv, const double mach, const double angle);
        // Определяет геометрические к-нты для заданного числа Маха и угла
        // нутации
        
        double CSV_GetMach(csv *p_csv); //Число маха
        double CSV_GetTren(csv *p_csv); //К-нт силы трения (Cx трен)
        double CSV_GetDon(csv *p_csv);  //К-нт донного сопротивления (Cx дон)
        double CSV_GetVoln(csv *p_csv); //К-нт волнового сопротивления (Cx волн)
        double CSV_GetPois(csv *p_csv); //К-нт сопротивления пояска (Cx пояс)
        double CSV_GetCn(csv *p_csv);   //К-нт нормальной силы (Cn)
        double CSV_GetCy(csv *p_csv);   //К-нт подъемной силы (Cy)
        double CSV_GetXcd(csv *p_csv);  //Положение центра давления от носика, м
        double CSV_GetCx(csv *p_csv);   //К-нт лобового сопротивления (Cx)
        double CSV_GetCyAl(csv *p_csv); //
        double CSV_GetCmAl(csv *p_csv); //
        double CSV_GetMzWz(csv *p_csv);  //
        double CSV_GetMxWx(csv *p_csv); //
        void CSV_Free(csv *p_csv);      //
        '''
    )

    lib = ffi.dlopen(LIB_FILE_NAME)

    return ffi, lib

FFI, KONTUR_LIB = load_lib()

class data(TypedDict):
    ng: float
    nh: float
    nk: float
    nfl: float
    d: float
    hb: float
    xct: float
    yint: float
    geometry_data: list


class KonturSolver:

    name = 'kontur_solver'

    preprocessed_data: data = dict(
        ng=None,
        nh=None,
        nk=None,
        nfl=None,
        d=None,
        hb=None,
        xct=None,
        yint=None,
        geometry_data=None
    )
    def preprocessor(self, data: dict, global_state: dict):

        self.preprocessed_data['ng'] = global_state['mcc_solver']['kontur']['ng']
        self.preprocessed_data['nh'] = global_state['mcc_solver']['kontur']['nh']
        self.preprocessed_data['nk'] = global_state['mcc_solver']['kontur']['nk']
        self.preprocessed_data['nfl'] = global_state['mcc_solver']['kontur']['nfl']
        self.preprocessed_data['d'] = global_state['mcc_solver']['kontur']['d']
        self.preprocessed_data['hb'] = global_state['mcc_solver']['kontur']['hb']
        self.preprocessed_data['xct'] = global_state['mcc_solver']['kontur']['xct']
        self.preprocessed_data['yint'] = global_state['mcc_solver']['kontur']['yint']
        self.preprocessed_data['geometry_data'] = global_state['mcc_solver']['kontur']['geometry_data']


    def run(self, data, global_state):
        p_data = self.preprocessed_data
        solver = KONTUR_LIB.CSV_New()
        KONTUR_LIB.CSV_SetNG(solver, p_data['ng'])
        KONTUR_LIB.CSV_SetNH(solver, p_data['nh'])
        KONTUR_LIB.CSV_SetNK(solver, p_data['nk'])
        KONTUR_LIB.CSV_SetNFL(solver, p_data['nfl'])
        KONTUR_LIB.CSV_SetDia(solver, p_data['d'])
        KONTUR_LIB.CSV_SetHb(solver, p_data['hb'])
        KONTUR_LIB.CSV_SetXct(solver, p_data['xct'])
        KONTUR_LIB.CSV_SetYInt(solver, p_data['yint'])

        for i, primitive in enumerate(p_data['geometry_data']):
            KONTUR_LIB.CSV_AddGeom(solver, i, primitive['type'], primitive['x1'], primitive['x2'],
                                   primitive['r1'], primitive['r2'], primitive['R'], primitive['n'])

        settings = data['settings']['kontur']

        attack_angle = np.linspace(settings['alpha_0'], settings['alpha_n'], settings['n_alpha'])
        mach_list = np.linspace(settings['mach_0'], settings['mach_n'], settings['n_mach'])

        mesh_tren = np.empty((attack_angle.shape[0], mach_list.shape[0]))
        mesh_don = np.empty_like(mesh_tren)
        mesh_voln = np.empty_like(mesh_tren)
        mesh_pois = np.empty_like(mesh_tren)
        mesh_cn = np.empty_like(mesh_tren)
        mesh_cy = np.empty_like(mesh_tren)
        mesh_xcd = np.empty_like(mesh_tren)
        mesh_cx = np.empty_like(mesh_tren)
        mesh_cyal = np.empty_like(mesh_tren)
        mesh_cmal = np.empty_like(mesh_tren)
        mesh_mxwx = np.empty_like(mesh_tren)
        mesh_mzwz = np.empty_like(mesh_tren)

        for i, atk_angl in enumerate(attack_angle):
            for j, mach in enumerate(mach_list):
                KONTUR_LIB.CSV_Solve(solver, mach, atk_angl)
                mesh_tren[i, j] = KONTUR_LIB.CSV_GetTren(solver)
                mesh_don[i, j] = KONTUR_LIB.CSV_GetDon(solver)
                mesh_voln[i, j] = KONTUR_LIB.CSV_GetVoln(solver)
                mesh_pois[i, j] = KONTUR_LIB.CSV_GetPois(solver)
                mesh_cn[i, j] = KONTUR_LIB.CSV_GetCn(solver)
                mesh_cy[i, j] = KONTUR_LIB.CSV_GetCy(solver)
                mesh_xcd[i, j] = KONTUR_LIB.CSV_GetXcd(solver)
                mesh_cx[i, j] = KONTUR_LIB.CSV_GetCx(solver)
                mesh_cyal[i, j] = KONTUR_LIB.CSV_GetCyAl(solver)
                mesh_cmal[i, j] = KONTUR_LIB.CSV_GetCmAl(solver)
                mesh_mxwx[i, j] = KONTUR_LIB.CSV_GetMxWx(solver)
                mesh_mzwz[i, j] = KONTUR_LIB.CSV_GetMzWz(solver)

        # global_state[KonturSolver.name] = dict(
        #     mesh_tren=mesh_tren,
        #     mesh_don=mesh_don,
        #     mesh_voln=mesh_voln,
        #     mesh_pois=mesh_pois,
        #     mesh_cn=mesh_cn,
        #     mesh_cy=mesh_cy,
        #     mesh_xcd=mesh_xcd,
        #     mesh_cx=mesh_cx,
        #     mesh_cyal=mesh_cyal,
        #     mesh_cmal=mesh_cmal,
        #     mesh_mxwx=mesh_mxwx,
        #     mesh_mzwz=mesh_mzwz
        # )
        global_state[KonturSolver.name] = dict(
            cx_list=mesh_cx[0],
            mach_list=mach_list
        )



if __name__ == '__main__':
    import matplotlib.pyplot as plt






    def print_solver(solver):
        print(
            KONTUR_LIB.CSV_GetMach(solver),
            KONTUR_LIB.CSV_GetTren(solver),
            KONTUR_LIB.CSV_GetDon(solver),
            KONTUR_LIB.CSV_GetVoln(solver),
            KONTUR_LIB.CSV_GetPois(solver),
            KONTUR_LIB.CSV_GetCn(solver),
            KONTUR_LIB.CSV_GetCy(solver),
            KONTUR_LIB.CSV_GetXcd(solver),
            KONTUR_LIB.CSV_GetCx(solver),
            KONTUR_LIB.CSV_GetCyAl(solver),
            KONTUR_LIB.CSV_GetCmAl(solver),
            KONTUR_LIB.CSV_GetMxWx(solver),
            KONTUR_LIB.CSV_GetMzWz(solver),
        )

    def test_kontur():
        solver = KONTUR_LIB.CSV_New()
        KONTUR_LIB.CSV_SetNG(solver, 4)
        KONTUR_LIB.CSV_SetNH(solver, 5)
        KONTUR_LIB.CSV_SetNK(solver, 7)
        KONTUR_LIB.CSV_SetNFL(solver, 2)
        KONTUR_LIB.CSV_SetDia(solver, 0.1524)
        KONTUR_LIB.CSV_SetHb(solver, 0.01 * 0.1524)
        KONTUR_LIB.CSV_SetXct(solver, 0.533)
        KONTUR_LIB.CSV_SetYInt(solver, 0.)
        KONTUR_LIB.CSV_AddGeom(solver, 0, KONTUR_LIB.G_CYLINDR, 0., .01, .01, .01, 0., 0.)
        KONTUR_LIB.CSV_AddGeom(solver, 1, KONTUR_LIB.G_CONE, .01, .06, .01, .02, 0., 0.)
        KONTUR_LIB.CSV_AddGeom(solver, 2, KONTUR_LIB.G_CONE, .06, .18, .02, .0425, 0., 0.)
        KONTUR_LIB.CSV_AddGeom(solver, 3, KONTUR_LIB.G_OGIVE, .18, .457, .0425, .0762, 3.22, 0.)
        KONTUR_LIB.CSV_AddGeom(solver, 4, KONTUR_LIB.G_CYLINDR, .457, .712, .0762, .0762, .0, 0.)
        KONTUR_LIB.CSV_AddGeom(solver, 5, KONTUR_LIB.G_CONE, .712, .723, .0762, .0748, .0, 0.)
        KONTUR_LIB.CSV_AddGeom(solver, 6, KONTUR_LIB.G_CONE, .723, .863, .0748, .06675, .0, 0.)

        attack_angle = np.linspace(0., 2., 5)
        mach_list = np.linspace(0.4, 3.1)

        mesh_tren = np.empty((attack_angle.shape[0], mach_list.shape[0]))
        mesh_don = np.empty_like(mesh_tren)
        mesh_voln = np.empty_like(mesh_tren)
        mesh_pois = np.empty_like(mesh_tren)
        mesh_cn = np.empty_like(mesh_tren)
        mesh_cy = np.empty_like(mesh_tren)
        mesh_xcd = np.empty_like(mesh_tren)
        mesh_cx = np.empty_like(mesh_tren)
        mesh_cyal = np.empty_like(mesh_tren)
        mesh_cmal = np.empty_like(mesh_tren)
        mesh_mxwx = np.empty_like(mesh_tren)
        mesh_mzwz = np.empty_like(mesh_tren)

        for i, atk_angl in enumerate(attack_angle):
            # print(f'Угол нутации: {atk_angl}')
            for j, mach in enumerate(mach_list):
                KONTUR_LIB.CSV_Solve(solver, mach, atk_angl)
                # KONTUR_LIB.CSV_GetMach(solver)
                mesh_tren[i, j] = KONTUR_LIB.CSV_GetTren(solver)
                mesh_don[i, j] = KONTUR_LIB.CSV_GetDon(solver)
                mesh_voln[i, j] = KONTUR_LIB.CSV_GetVoln(solver)
                mesh_pois[i, j] = KONTUR_LIB.CSV_GetPois(solver)
                mesh_cn[i, j] = KONTUR_LIB.CSV_GetCn(solver)
                mesh_cy[i, j] = KONTUR_LIB.CSV_GetCy(solver)
                mesh_xcd[i, j] = KONTUR_LIB.CSV_GetXcd(solver)
                mesh_cx[i, j] = KONTUR_LIB.CSV_GetCx(solver)
                mesh_cyal[i, j] = KONTUR_LIB.CSV_GetCyAl(solver)
                mesh_cmal[i, j] = KONTUR_LIB.CSV_GetCmAl(solver)
                mesh_mxwx[i, j] = KONTUR_LIB.CSV_GetMxWx(solver)
                mesh_mzwz[i, j] = KONTUR_LIB.CSV_GetMzWz(solver)

        plt.plot(mach_list, mesh_cx[0], label='alpha 0.')
        plt.plot(mach_list, mesh_cx[1], label='alpha 1.')
        plt.plot(mach_list, mesh_cx[2], label='alpha 1.5')
        plt.plot(mach_list, mesh_cx[3], label='alpha 2.')
        plt.legend()
        plt.show()

    test_kontur()
